% This function computes supplementary matrices that are used by JointPopDensEstLogLikeKnownTimes
% and JointPopDensEstLogLikeKnownDistribution functions in computing the
% loglikelihood of the joint population and size estimation model in [1].
% allIndTau, discIndTau and likeConst are used by
% JointPopDensEstLogLikeKnownTimes when the actual arrival/departure times of the
% discovered individuals are known.
% allIndTau and Psi are used by JointPopDensEstLogLikeKnownDistribution
% when only the arrival/departure time distribution of the individuals is
% known, but not the actual times.
% NOTE: For the latter case, discIndEntDepTimes passed as argument should be empty.
%
% Input:
% agentTraj is a cell containing M matrices where each Matrix represents
% the trajectory of an agent. Each matrix has three columns where the first
% column is the arrival time of the agent to a location, where the second
% column is the departure time of the agent from that location, and where the
% third column is the ID of that location.
% contMat is an SxMxK matrix where element ijk is equal to the number of times
% that agent j contacts individual i in location k, for 1<=i<=S, 1<=j<=M,
% and 1 <= k <= K.
% allEntDepTimes is an Nx2 containing the arrival and departure times of all
% the N individuals to/from the event.
% discIndEntDepTimes is an Sx2 matrix containing the arrival and departure
% times of all the S discovered individuals to/from the event.
%
% Output:
% allIndTau is a matrix where element ki is equal to the summation of the overlap
% times of each agent's presence in location k and the presence of the i'th individual 
% (from the sample of the individual's arrival/departure times from allEntDepTimes) in the event. 
% The matrix has K rows. The number of columns is equal to the samples of the individuals' arrival/departure
% times from allEntDepTimes.
% discIndTau is a KxS matrix, where element ki is equal to the summation of the overlap
% times of each agent's presence in location k and the presence of the i'th
% discovered individual in the event.
% likeConst is a constant part of the likelihood function given in Section II-A of [2] (for more
% information refer to the function JointPopDensEstLogLikeKnownTimes).
% Psi is used to compute a part of the loglikelihood function given in Equation (17) of
% [1]. In particular it is a matrix with S rows. The number of columns is equal to the
% number of samples of the arrival/departure times. It contains
% sum_{l=1}^{K}sum_{j=1}^{M}log(k_{ij}^{(l)})t_{at, dt}^{j,(l)} where 1 <=i <=S
% is the id of the discovered individual and (at, dt) is a sample
% arrival/departure time.
%
% [1]: "Opportunistic Sampling for Joint Population Size and Density Estimation",
% F. M. Naini, O. Dousse, P. Thiran, and M. Vetterli, IEEE Transactions on Mobile
% Computing, Volume 14, Issue 12, 2015, available at: http://infoscience.epfl.ch/ and http://rr.epfl.ch/

function [allIndTau, discIndTau, likeConst, Psi] = SuppMatCalc(agentTraj, contMat, allEntDepTimes, discIndEntDepTimes)

numSamples = size(allEntDepTimes, 1);
[S, M, K] = size(contMat);
allIndTau = zeros(K, numSamples);
Psi = zeros(S, numSamples);
discIndTau = zeros(K, S);
likeConst = 0;

% concatenating all the agents' traces into one matrix
agentTrajMat = [];
for agentIter = 1:M
    traj = agentTraj{agentIter};
    agentTrajMat = [agentTrajMat; [repmat(agentIter, [size(traj, 1), 1]), traj]];
end

%% computing Psi matrix
if isempty(discIndEntDepTimes)
    % in this case, discIndEntDepTimes is empty, so the discovered
    % individuals' actual arrival departure times is not known. Hence we need the
    % Psi matrix for computing the loglikelihood by the
    % JointPopDensEstLogLikeKnownDistribution function
    fprintf('\nComputing the Psi matrix:');
    for discIndIter = 1:S
        % iteration over the discovered individuals
        % discIndIter
        if mod(discIndIter, round(S * 0.05)) == 0
            fprintf('\n%d percent done.', round(discIndIter / S * 100));
        end
        for indIter = 1:numSamples
            % iteration over the samples of the individuals arrival/departure times
   
            % computing t_{at,dt}^{j,(l)}
            overlapTimeIndAgenTrans = max(0, min(agentTrajMat(:, 3), allEntDepTimes(indIter, 2)) - max(agentTrajMat(:, 2), allEntDepTimes(indIter, 1)));
            overlapTimeMat = sparse(agentTrajMat(:, 1), agentTrajMat(:, 4), overlapTimeIndAgenTrans, M, K);
         
            contMatTemp = reshape(contMat(discIndIter, :, :), [M, K]);
            PsiPart = contMatTemp .* log(overlapTimeMat);
            PsiPart(contMatTemp == 0) = 0;
            Psi(discIndIter, indIter) = sum(sum(PsiPart));
        end
    end
end

%% computing likeConst
% this value is used by the JointPopDensEstLogLikeKnownTimes function
if ~isempty(discIndEntDepTimes)
    for discIndIter = 1:S
        % iteration over the discovered individuals

        overlapTimeIndAgenTrans = max(0, min(agentTrajMat(:, 3), discIndEntDepTimes(discIndIter, 2)) - max(agentTrajMat(:, 2), discIndEntDepTimes(discIndIter, 1)));
        overlapTimeMat = sparse(agentTrajMat(:, 1), agentTrajMat(:, 4), overlapTimeIndAgenTrans, M, K);

        contMatTemp = reshape(contMat(discIndIter, :, :), [M, K]);
        likeConstPart = contMatTemp .* log(overlapTimeMat);
        likeConstPart(contMatTemp == 0) = 0;
        likeConst = likeConst + sum(sum(likeConstPart));
    end
end

%% computing tau matrices
for agentIter = 1:M
    % iteration over agents
    traj = agentTraj{agentIter};
    for locIter = 1:size(traj, 1)
     
        visitedLoc = traj(locIter, 3);
        locEntTime = traj(locIter, 1);
        locDepTime = traj(locIter, 2);
     
        %computing the overlap time between current stay and the
        %individual
        overlapTimeIndAgenTrans = max(0, min(locDepTime, allEntDepTimes(:, 2)) - max(locEntTime, allEntDepTimes(:, 1)));
     
        % updating Tau matrix
        allIndTau(visitedLoc, :) = allIndTau(visitedLoc, :) + overlapTimeIndAgenTrans';
     
        %computing the overlap time between current stay and the
        %individual
        if ~isempty(discIndEntDepTimes)
            overlapTimeDiscIndAgenTrans = max(0, min(locDepTime, discIndEntDepTimes(:, 2)) - max(locEntTime, discIndEntDepTimes(:, 1)));
         
            % updating discIndTau matrix
            discIndTau(visitedLoc, :) = discIndTau(visitedLoc, :) + overlapTimeDiscIndAgenTrans';
         
        end
    end
end