% This file demonstrates the population size estimation model described in
% [1]. Traces are first generated and then the contact matrix is
% computed. Afterwards population size is estimated by varying the amount of
% information in the traces: the number of agents and observation window are changed.
%
% Similar methodology is used in producing Figure 5 and Section 5.3 of [1], and in
% producing Figure 2, 6 and 7 and Section I-C of [2].
%
% [1]: "Opportunistic Sampling for Joint Population Size and Density Estimation",
% F. M. Naini, O. Dousse, P. Thiran, and M. Vetterli, IEEE Transactions on Mobile
% Computing, Volume 14, Issue 12, 2015, available at: http://infoscience.epfl.ch/ and http://rr.epfl.ch/
% [2]: "Opportunistic Sampling for Joint Population Size and Density
% Estimation: Supplementary Notes", F. M. Naini, O. Dousse, P. Thiran,
% and M. Vetterli, 2015, available at: http://infoscience.epfl.ch/ and http://rr.epfl.ch/

clear all;
close all;
clc;

%% selecting parameters
% Feel free to modify these values
N = 1000; % population size (i.e., number of individuals)
M = 10; % number of agents
alpha0 = 1.5; % parameter of the gamma distribution modeling the contact rates
beta0 = 1000; % parameter of the gamma distribution modeling the contact rates
eventDuration = 600; % duration of the experiment (e.g., event duration) or observation window
% optimization options
options = optimset('GradObj', 'off', 'Display', 'off', 'MaxFunEvals', 300000, 'Algorithm', 'interior-point', 'MaxIter', 300000, 'TolX', 1e-12, 'TolCon', 1e-12, 'TolFun', 1e-9);

%% simulating agents' and individuals' contact traces
% In the following function, contTrace is a cell containing M matrices where
% each matrix corresponds to the traces of one agent.
% Each matrix has two columns where the first column
% is the ID of the individual contacted by the agent and the second column
% is the time of contact. The IDs of individuals are numbered from 1 to S
% where S <= N is the total number of discovered individuals. The time of contacts are
% between 0 and eventDuration.
% agentEntDepTimes is an Mx2 matrix where the first column is the arrival
% times of the agents to the event and the second column is the departure
% times of the agents from the event.
% allEntDepTimes is an Nx2 containing the arrival and departure times of all
% the N individuals to/from the event.
% discIndEntDepTimes is an Sx2 matrix containing the arrival and departure
% times of all the S discovered individuals to/from the event.
fprintf('\nGenerating traces:\n');
[contTrace, S, agentEntDepTimes, allEntDepTimes, discIndEntDepTimes] = traceSimulate1(alpha0, beta0, eventDuration, M, N);

%% obtaining the contact matrix from the agents' and individuals' contact traces
% In the following function, contMat is an SxM matrix where element ij is equal to the number of times
% that agent j contacts individual i. 1 <= i <= S and 1 <= j <= M.
[contMat] = contMatCalc(contTrace, M, S);

%% displaying some properties of the traces
% displaing on the console
fprintf('Number of individuals N: %d \n', N);
fprintf('Number of agents M: %d\n', M);
fprintf('Duration of event (observation window): %d\n', eventDuration);
fprintf('Parameters of the gamma distribution, alpha: %d, beta: %d\n', alpha0, beta0);
fprintf('Number of discovered individuals S: %d (%1.1d%% coverage)\n', S, S / N * 100);

% estimating initial alpha0 and beta0 for optimization
[alphaStart, betaStart] = parameterStart1(contMat, agentEntDepTimes, allEntDepTimes);

%% Estimating the population size by varying the number of agents
fprintf('\nEstimating the population size by using a subset of information: reducing the number of agents\n');
agentsValues = floor(M / 2) + 2:1:M;    % feel free to change these values
agentsResult = zeros(length(agentsValues), 3);
SValues = zeros(length(agentsValues), 1);
for agentNumberIter = 1:length(agentsValues) 
    % going over all the possible subsets of agents
    fprintf('Shrinking down the number of agents to: %d\n', agentsValues(agentNumberIter));
    agentsSubsets = nchoosek([1:M], agentsValues(agentNumberIter));
    NHatMat = zeros(size(agentsSubsets, 1), 1);
    SValue = zeros(size(agentsSubsets, 1), 1);
    for windowIter = 1:size(agentsSubsets, 1)
        % shrinking the agents' traces to current number of agents
        [contTraceShrink, SShrink, agentEntDepTimesShrink, allEntDepTimesShrink, discIndEntDepTimesShrink] = traceShrinker(contTrace, agentEntDepTimes, allEntDepTimes, discIndEntDepTimes, agentsSubsets(windowIter, :), [0 eventDuration]);
     
        % estimating the population size
        [contMatShrink] = contMatCalc(contTraceShrink, agentsValues(agentNumberIter), SShrink);
     
        % estimating initial alpha0 and beta0 for optimization
        %[alphaHat, betaHat] = parameterStart1(contMatShrink, agentEntDepTimesShrink, allEntDepTimesShrink);
     
        % performing the optimization
        funcOpt = @(parameters)PopEstLogLikeKnownTimes(parameters, contMatShrink, discIndEntDepTimesShrink, agentEntDepTimesShrink, allEntDepTimesShrink);
        params1 = fmincon(funcOpt, [alphaStart betaStart], [], [], [], [], [0 0], [Inf Inf], [], options);
        [~, NHat1] = funcOpt(params1);
        NHatMat(windowIter) = NHat1;
        SValue(windowIter) = SShrink;
    end
 
    %saving the mean and standard deviation of all the estimated population
    %sizes with the current number of agents
    agentsResult(agentNumberIter, 1) = mean(NHatMat);
    agentsResult(agentNumberIter, [2 3]) = myconfintcal1(NHatMat, .9);
    SValues(agentNumberIter) = mean(SValue);
end

% plotting results
figure, errorbar(agentsValues, agentsResult(:, 1), agentsResult(:, 1) - agentsResult(:, 2), agentsResult(:, 3) - agentsResult(:, 1), 'linewidth', 2);
hold on;
plot(agentsValues, N * ones(length(agentsValues), 1), 'r--', 'linewidth', 2);
title('Population size estimation when varying the number of agents');
xlabel('Number of agents');
legend('Average and 90% percentile', 'Ground-truth', 'Location', 'southeast');

figure, plot(agentsValues, SValues, '.-', 'linewidth', 2, 'markersize', 10);
hold on;
plot(agentsValues, N * ones(length(agentsValues), 1), 'r--', 'linewidth', 2);
title('Number of discovered individuals S when varying the number of agents');
xlabel('Number of agents');
ylim([min(SValues) * .9 N * 1.1]);
legend('Average of S', 'Population size ground-truth', 'Location', 'southeast');

%% Estimating the population size by varying the time duration of the event
fprintf('\nEstimating the population size by using a subset of information: reducing the observation window\n');
windowStep = floor(eventDuration / 6);
windowValues = floor(eventDuration / 2):windowStep:eventDuration; % feel free to change these values
windowResult = zeros(length(windowValues), 3);
for windowNumberIter = 1:length(windowValues) 
    % going over all the possible subsets of agents
    fprintf('Shrinking down the observation window size to: %d\n', windowValues(windowNumberIter));
    windowSubsets = [0:windowStep:eventDuration - windowValues(windowNumberIter)];
    NHatMat = zeros(length(windowSubsets), 1);
    SValue = zeros(size(windowSubsets, 1), 1);
    for windowIter = 1:length(windowSubsets)
        % shrinking the agents' traces to current number of agents
        [contTraceShrink, SShrink, agentEntDepTimesShrink, allEntDepTimesShrink, discIndEntDepTimesShrink] = traceShrinker(contTrace, agentEntDepTimes, allEntDepTimes, discIndEntDepTimes, [1:M], [windowSubsets(windowIter) windowSubsets(windowIter) + windowValues(windowNumberIter)]);
     
        % estimating the population size
        [contMatShrink] = contMatCalc(contTraceShrink, M, SShrink);
             
        % estimating initial alpha0 and beta0 for optimization
        %[alphaHat, betaHat] = parameterStart1(contMatShrink, agentEntDepTimesShrink, allEntDepTimesShrink);
     
        % performing the optimization
        funcOpt = @(parameters)PopEstLogLikeKnownTimes(parameters, contMatShrink, discIndEntDepTimesShrink, agentEntDepTimesShrink, allEntDepTimesShrink);
        params1 = fmincon(funcOpt, [alphaStart betaStart], [], [], [], [], [0 0], [Inf Inf], [], options);
        [~, NHat1] = funcOpt(params1);
        NHatMat(windowIter) = NHat1;
        SValue(windowIter) = SShrink;
    end
 
    %saving the mean and standard deviation of all the estimated population
    %sizes with the current number of agents
    windowResult(windowNumberIter, 1) = mean(NHatMat);
    windowResult(windowNumberIter, [2 3]) = myconfintcal1(NHatMat, .9);
    SValues(windowNumberIter) = mean(SValue);
end

% plotting results
figure, errorbar(windowValues, windowResult(:, 1), windowResult(:, 1) - windowResult(:, 2), windowResult(:, 3) - windowResult(:, 1), 'linewidth', 2);
hold on;
plot(windowValues, N * ones(length(agentsValues), 1), 'r--', 'linewidth', 2);
title('Population size estimation when varying the observation window');
xlabel('Observation window length');
legend('Average and 90% percentile', 'Ground-truth', 'Location', 'southeast');

figure, plot(windowValues, SValues, '.-', 'linewidth', 2, 'markersize', 10);
hold on;
plot(windowValues, N * ones(length(windowValues), 1), 'r--', 'linewidth', 2);
title('Number of discovered individuals S when varying the observation window');
xlabel('Observation window length');
legend('Average of S', 'Population size ground-truth', 'Location', 'southeast');
ylim([min(SValues) * .9 N * 1.1]);