% This file demonstrates the joint population size and density estimation model
% described in [1]. Traces are first generated and then the contact matrix is
% computed. Afterwards population size and density is jointly estimated by
% using the actual arrival/departure times of the individuals, the actual
% arrival/departure time distribution of the individuals, and the approximated
% arrival/departure time distribution of individuals.
%
% Similar methodology is used in Figure 6 and Section 7.2 of [1].
%
% [1]: "Opportunistic Sampling for Joint Population Size and Density Estimation",
% F. M. Naini, O. Dousse, P. Thiran, and M. Vetterli, IEEE Transactions on Mobile
% Computing, Volume 14, Issue 12, 2015, available at: http://infoscience.epfl.ch/ and http://rr.epfl.ch/

clear all;
close all;
clc;

%% selecting parameters
% Feel free to modify these values
N = 1000; % population size (i.e., number of individuals)
M = 10; % number of agents
K = 32; % number of locations -- this value might change by the traceSimulate2 function
alpha0 = 1.5; %parameter of the gamma distribution modeling the contact rates
beta0 = 1000 / K; %parameter of the gamma distribution modeling the contact rates
eventDuration = 600; %duration of the experiment (e.g., event duration) or observation window
% optimization options
options = optimset('GradObj', 'off', 'Display', 'iter', 'MaxFunEvals', 300000, 'Algorithm', 'interior-point', 'MaxIter', 300000, 'TolX', 1e-12, 'TolCon', 1e-12, 'TolFun', 1e-9);

%% simulating agents' and individuals' contact traces
% In the following function, contTrace is a cell containing M matrices where
% each matrix corresponds to the traces of one agent.
% Each matrix has three columns where the first column
% is the ID of the individual contacted by the agent, where the second column
% is the time of contact and where the third columns is the location where the contact
% occurred. The IDs of individuals are numbered from 1 to S
% where S <= N is the total number of discovered individuals. The time of contacts are
% between 0 and eventDuration. The locations of contacts are between 1 and K.
% agentEntDepTimes is an Mx2 matrix where the first column is the arrival
% times of the agents to the event and the second column is the departure
% times of the agents from the event.
% agentTraj is a cell containing M matrices where each Matrix represents
% the trajectory of an agent. Each matrix has three columns where the first
% column is the arrival time of the agent to a location, where the second
% column is the departure time of the agent from that location, and where the
% third column is the ID of that location.
% allEntDepTimes is an Nx2 containing the arrival and departure times of all
% the N individuals to/from the event.
% discIndEntDepTimes is an Sx2 matrix containing the arrival and departure
% times of all the S discovered individuals to/from the event.
% PiVec is a vector containing the density values of each location. This
% vector sums to one and is based on the definition in Section 6.1 of [1].
fprintf('\nGenerating traces:\n');
[contTrace, S, agentEntDepTimes, agentTraj, allEntDepTimes, discIndEntDepTimes, K, PiVec] = traceSimulate2(alpha0, beta0, eventDuration, M, N, K);

%% obtaining the contact matrix from the agents' and individuals' contact traces
% contMat is an SxMxK matrix where element ijk is equal to the number of times
% that agent j contacts individual i in location k, for 1 <= i <= S,  1 <= j <= M,
% and 1 <= k < K.
[contMat] = contMatCalc2(contTrace, M, S, K);

%% displaying some properties of the traces
% plotting individuals' arrival/departure times -- scatter plot
figure, plot(allEntDepTimes(:, 1), allEntDepTimes(:, 2), '.', 'markersize', 5);
hold on;
xlim([0 eventDuration]);
ylim([0 eventDuration])
xlabel('Arrival time')
ylabel('Departure time')
title('Scatter plot of the arrival/departure times')
% plotting agents' arrival/departure times
plot(agentEntDepTimes(:, 1), agentEntDepTimes(:, 2), 'r*', 'markersize', 3);
legend('Individuals', 'Agents', 'Location', 'southeast');

% plotting individuals' arrival/departure times -- density function
[bandwidth, density, X, Y] = kde2d(allEntDepTimes, 2 ^ 8, [0 0], [eventDuration eventDuration]);
figure, imagesc(flipud(density));
cp = colormap('gray');
colormap(flipud(cp));
ax = gca;
ax.XTick = [0 64 128 192 256];
ax.XTickLabel = {'0' num2str(eventDuration / 4) num2str(eventDuration / 2) num2str(eventDuration * 3 / 4) num2str(eventDuration)};
ax.YTick = [0 64 128 192 256];
ax.YTickLabel = fliplr({'0' num2str(eventDuration / 4) num2str(eventDuration / 2) num2str(eventDuration * 3 / 4) num2str(eventDuration)});
xlabel('Arrival time');
ylabel('Departure time');
title('Joint arrival/departure time distribution of the individuals');
hold on;
plot(agentEntDepTimes(:, 1), agentEntDepTimes(:, 2), 'r*', 'markersize', 3);

% plotting the marginal distributions
[fArv, xiArv] = ksdensity(allEntDepTimes(:, 1), 'support', [- 1 eventDuration + 1], 'kernel', 'triangle');
[fDep, xiDep] = ksdensity(allEntDepTimes(:, 2), 'support', [- 1 eventDuration + 1], 'kernel', 'triangle');
[fDur, xiDur] = ksdensity(allEntDepTimes(:, 2) - allEntDepTimes(:, 1), 'support', [- 1 eventDuration + 1], 'kernel', 'triangle');
figure, subplot 311
plot(xiArv, fArv, 'markersize', 2);
title('Marginal distributions: Arrival times');
xlim([- 1 eventDuration + 1]);
subplot 312
plot(xiDep, fDep, 'markersize', 2);
title('Departure times');
xlim([- 1 eventDuration + 1]);
subplot 313
plot(xiDur, fDur, 'markersize', 2);
title('Duration of stay');
xlabel('Time');
xlim([- 1 eventDuration + 1]);

% plotting the evolution of the number of discovered individuals by agents,
% and the total number of discovered individuals and the duration of stay for
% each agent
displayTrace(contTrace, eventDuration, M, agentEntDepTimes);

% displaying on the console
fprintf('Number of individuals N: %d \n', N);
fprintf('Number of agents M: %d\n', M);
fprintf('Number of locations K: %d\n', K);
fprintf('Duration of event (observation window): %d\n', eventDuration);
fprintf('Parameters of the gamma distribution, alpha: %d, beta: %d\n', alpha0, beta0);
fprintf('Number of discovered individuals S: %d (%1.1d%% coverage)\n', S, S / N * 100);


%% Jointly Estimating the population size and density by using the actual
% arrival/departure times of the discovered individuals

fprintf('\nJointly estimating the population size and density by using the actual arrival/departure times of the discovered individuals:\n');

% estimating initial alpha0 and beta0 for optimization
[alphaStart, betaStart] = parameterStart1(sum(contMat, 3), agentEntDepTimes, allEntDepTimes);

% computing the Tau matrices, which are used in likelihood computation
[allIndTau, discIndTau, likeConst, ~] = SuppMatCalc(agentTraj, contMat, allEntDepTimes, discIndEntDepTimes);

% focusing on the locations in which at least one individual has been
% contacted by an agent
visitedLocList = find(reshape(sum(sum(contMat, 1), 2), [1, K]) > 0);
allIndTau = allIndTau(visitedLocList, :);
if ~ isempty(discIndTau)
    discIndTau = discIndTau(visitedLocList, :);
end

% performing the optimization
funcOpt = @(parameters)JointPopDensEstLogLikeKnownTimes(parameters, allIndTau, discIndTau, contMat(:, :, visitedLocList), likeConst);
PiVecStart = reshape(sum(sum(contMat(:, :, visitedLocList), 1), 2), [1, length(visitedLocList)]);
PiVecStart = PiVecStart ./ sum(allIndTau, 2)';
PiVecStart = PiVecStart / sum(PiVecStart);
paramsStart = [alphaStart * PiVecStart betaStart / length(visitedLocList)];
lb = 0 * ones(length(visitedLocList) + 1, 1);
lb(end) = paramsStart(end) * .1;
ub = Inf * ones(length(visitedLocList) + 1, 1);
ub(end) = paramsStart(end) * 10;
params1 = fmincon(funcOpt, paramsStart, [], [], [], [], lb', ub', [], options);
%  estimated population size
[~, NHat1] = funcOpt(params1);
% estimating density in vector form (i.e., 1D)
PiVecHat1 = zeros(1, K);
PiVecHat1(visitedLocList) = params1(1:end - 1) / sum(params1(1:end - 1));

% changing the estimated density into two dimensions
% grid sizes are defined in the traceSimulate2 function
xGrids = fix(sqrt(K / 2) * 2); % number of locations in the x-axis
yGrids = fix(sqrt(K / 2)); % number of locations in the y-axis
PiMat = reshape(PiVec, yGrids, xGrids);
PiMatHat1 = reshape(PiVecHat1, yGrids, xGrids);

% smoothing the estimated density
[X1, X2] = meshgrid([0:fix(sqrt(K) / 2)], [0:fix(sqrt(K) / 2)]);
sigmaScale = K ^ (0.5) / 20;
miu1 = [fix(sqrt(K) / 2) / 2 fix(sqrt(K) / 2) / 2];
Sigma1 = [1 0; 0 1] * sigmaScale;
h = mvnpdf([X1(:) X2(:)], miu1, Sigma1);
h = reshape(h, fix(sqrt(K) / 2) + 1, fix(sqrt(K) / 2) + 1);
h = h / sum(sum(h));

PiMatHatSmooth1 = filter2(h, PiMatHat1, 'same');

% plotting the results
figure, surf([0:xGrids - 1], [0:yGrids - 1], PiMat);
xlabel('x-axis');
ylabel('y-axis');
title('Density ground-truth');

% figure, surf([0:xGrids - 1], [0:yGrids - 1], PiMatHat1);
% xlabel('x-axis');
% ylabel('y-axis');
% title('Estimated density by using actual arrival/departure times');

figure, surf([0:xGrids - 1], [0:yGrids - 1], PiMatHatSmooth1);
xlabel('x-axis'); 
ylabel('y-axis');
title('Smoothed estimated density by using actual arrival/departure times');

fprintf('Estimated population size: %d\n', NHat1);
fprintf('Estimated parameters of the gamma distribution, alpha: %d, beta: %d\n', sum(params1(1:end - 1)), params1(end));

%% Jointly estimating the population size and density by using the actual
% arrival/departure time distribution of the discovered individuals

fprintf('\nJointly estimating the population size and density by using the actual arrival/departure time distribution of the individuals:\n');

% estimating initial alpha0 and beta0 for optimization
[alphaStart, betaStart] = parameterStart1(sum(contMat, 3), agentEntDepTimes, allEntDepTimes);

% computing the Tau and Psi matrices, which are used in likelihood computation
[allIndTau, ~, ~, Psi] = SuppMatCalc(agentTraj, contMat, allEntDepTimes, []);

% focusing on the locations in which at least one individual has been
% contacted by an agent
visitedLocList = find(reshape(sum(sum(contMat, 1), 2), [1, K]) > 0);
allIndTau = allIndTau(visitedLocList, :);

% performing the optimization
funcOpt = @(parameters)JointPopDensEstLogLikeKnownDistribution(parameters, allIndTau, contMat(:, :, visitedLocList), Psi);
PiVecStart = reshape(sum(sum(contMat(:, :, visitedLocList), 1), 2), [1, length(visitedLocList)]);
PiVecStart = PiVecStart ./ sum(allIndTau, 2)';
PiVecStart = PiVecStart / sum(PiVecStart);
paramsStart = [alphaStart * PiVecStart betaStart / length(visitedLocList)];
lb = 0 * ones(length(visitedLocList) + 1, 1);
lb(end) = paramsStart(end) * .1;
ub = Inf * ones(length(visitedLocList) + 1, 1);
ub(end) = paramsStart(end) * 10;
params2 = fmincon(funcOpt, paramsStart, [], [], [], [], lb', ub', [], options);
%  estimated population size
[~, NHat2] = funcOpt(params2);
% estimating density i 1D form
PiVecHat2 = zeros(1, K);
PiVecHat2(visitedLocList) = params2(1:end - 1) / sum(params2(1:end - 1));

% changing the estimated density into two dimensions
% grid sizes are defined in the traceSimulate2 function
xGrids = fix(sqrt(K / 2) * 2); % number of locations in the x-axis
yGrids = fix(sqrt(K / 2)); % number of locations in the y-axis
PiMat = reshape(PiVec, yGrids, xGrids);
PiMatHat2 = reshape(PiVecHat2, yGrids, xGrids);

% smoothing the estimated density
[X1, X2] = meshgrid([0:fix(sqrt(K) / 2)], [0:fix(sqrt(K) / 2)]);
sigmaScale = K ^ (0.5) / 20;
miu1 = [fix(sqrt(K) / 2) / 2 fix(sqrt(K) / 2) / 2];
Sigma1 = [1 0; 0 1] * sigmaScale;
h = mvnpdf([X1(:) X2(:)], miu1, Sigma1);
h = reshape(h, fix(sqrt(K) / 2) + 1, fix(sqrt(K) / 2) + 1);
h = h / sum(sum(h));

PiMatHatSmooth2 = filter2(h, PiMatHat2, 'same');

% plotting the results
% figure, surf([0:xGrids - 1], [0:yGrids - 1], PiMat);
% xlabel('x-axis');
% ylabel('y-axis');
% title('Density ground-truth');

% figure, surf([0:xGrids - 1], [0:yGrids - 1], PiMatHat2);
% xlabel('x-axis');
% ylabel('y-axis');
% title('Estimated density by using actual arrival/departure time distribution');

figure, surf([0:xGrids - 1], [0:yGrids - 1], PiMatHatSmooth2);
xlabel('x-axis');
ylabel('y-axis');
title('Smoothed estimated density by using actual arrival/departure time distribution');

fprintf('Estimated population size: %d\n', NHat2);
fprintf('Estimated parameters of the gamma distribution, alpha: %d, beta: %d\n', sum(params2(1:end - 1)), params2(end));

%% Jointly estimating the population size and density by using the approximated
% arrival/departure time distribution of the discovered individuals
% Using Gaussian approximation of the individuals' arrival/departure times distribution

fprintf('\nJointly estimating the population size and density by using the Gaussian approximation of arrival/departure time distribution of the individuals:\n');

% Gaussian approximation of the distribution
numSammples = S; % numSamples is the number of samples of the distribution over which
% we compute the expectation in the likelihood function.
allEntDepTimesGauEst(:, 1) = randn(numSammples, 1) * std(allEntDepTimes(:, 1)) + mean(allEntDepTimes(:, 1));
allEntDepTimesGauEst(:, 2) = randn(numSammples, 1) * std(allEntDepTimes(:, 2)) + mean(allEntDepTimes(:, 2));
allEntDepTimesGauEst(allEntDepTimesGauEst(:, 1) < 0, 1) = 0;
allEntDepTimesGauEst(allEntDepTimesGauEst(:, 2) > eventDuration, 2) = eventDuration;
allEntDepTimesGauEst(allEntDepTimesGauEst(:, 2) < 0, 2) = 0;
allEntDepTimesGauEst(allEntDepTimesGauEst(:, 1) > eventDuration, 1) = eventDuration;
allEntDepTimesGauEst = sort(allEntDepTimesGauEst, 2);

% estimating initial alpha0 and beta0 for optimization
[alphaStart, betaStart] = parameterStart1(sum(contMat, 3), agentEntDepTimes, allEntDepTimes);

% computing the Tau and Psi matrices, which are used in likelihood computation
[allIndTau, ~, ~, Psi] = SuppMatCalc(agentTraj, contMat, allEntDepTimesGauEst, []);

% focusing on the locations in which at least one individual has been
% contacted by an agent
visitedLocList = find(reshape(sum(sum(contMat, 1), 2), [1, K]) > 0);
allIndTau = allIndTau(visitedLocList, :);

% performing the optimization
funcOpt = @(parameters)JointPopDensEstLogLikeKnownDistribution(parameters, allIndTau, contMat(:, :, visitedLocList), Psi);
PiVecStart = reshape(sum(sum(contMat(:, :, visitedLocList), 1), 2), [1, length(visitedLocList)]);
PiVecStart = PiVecStart ./ sum(allIndTau, 2)';
PiVecStart = PiVecStart / sum(PiVecStart);
paramsStart = [alphaStart * PiVecStart betaStart / length(visitedLocList)];
lb = 0 * ones(length(visitedLocList) + 1, 1);
lb(end) = paramsStart(end) * .1;
ub = Inf * ones(length(visitedLocList) + 1, 1);
ub(end) = paramsStart(end) * 10;
params3 = fmincon(funcOpt, paramsStart, [], [], [], [], lb', ub', [], options);
%  estimated population size
[~, NHat3] = funcOpt(params3);
% estimating density i 1D form
PiVecHat3 = zeros(1, K);
PiVecHat3(visitedLocList) = params3(1:end - 1) / sum(params3(1:end - 1));

% changing the estimated density into two dimensions
% grid sizes are defined in traceSimulate2 function
xGrids = fix(sqrt(K / 2) * 2); % number of locations in the x-axis
yGrids = fix(sqrt(K / 2)); % number of locations in the y-axis
PiMat = reshape(PiVec, yGrids, xGrids);
PiMatHat3 = reshape(PiVecHat3, yGrids, xGrids);

% smoothing the estimated density
[X1, X2] = meshgrid([0:fix(sqrt(K) / 2)], [0:fix(sqrt(K) / 2)]);
sigmaScale = K ^ (0.5) / 20;
miu1 = [fix(sqrt(K) / 2) / 2 fix(sqrt(K) / 2) / 2];
Sigma1 = [1 0; 0 1] * sigmaScale;
h = mvnpdf([X1(:) X2(:)], miu1, Sigma1);
h = reshape(h, fix(sqrt(K) / 2) + 1, fix(sqrt(K) / 2) + 1);
h = h / sum(sum(h));

PiMatHatSmooth3 = filter2(h, PiMatHat3, 'same');

% plotting the results
% figure, surf([0:xGrids - 1], [0:yGrids - 1], PiMat);
% xlabel('x-axis');
% ylabel('y-axis');
% title('Density ground-truth');

% figure, surf([0:xGrids - 1], [0:yGrids - 1], PiMatHat3);
% xlabel('x-axis');
% ylabel('y-axis');
% title('Estimated density by using Gaussian approximation of arrival/departure time distribution');

figure, surf([0:xGrids - 1], [0:yGrids - 1], PiMatHatSmooth3);
xlabel('x-axis');
ylabel('y-axis');
title('Smoothed estimated density by using Gaussian approximation of arrival/departure time distribution');

fprintf('Estimated population size: %d\n', NHat3);
fprintf('Estimated parameters of the gamma distribution, alpha: %d, beta: %d\n', sum(params3(1:end - 1)), params3(end));